--[[
    https://us.forums.blizzard.com/en/wow/t/wow-classic-ui-api-change-for-unithealth/446596
    https://www.townlong-yak.com/framexml/37176/GlobalStrings.lua/RU
    https://github.com/tomrus88/BlizzardInterfaceCode/blob/master/Interface/SharedXML/SharedColorConstants.lua#L69-L78

    Multiplication factors prefix symbol:
    1e+24 1,000,000,000,000,000,000,000,000 yotta Y
    1e+21 1,000,000,000,000,000,000,000 zetta Z
    1e+18 1,000,000,000,000,000,000 exa E
    1e+15 1,000,000,000,000,000 peta P
    1e+12 1,000,000,000,000 tera T
    1e+09 1,000,000,000 giga G
    1e+06 1,000,000 mega M
    1e+03 1,000 kilo k
    1e+02 100 hecto h
    1e+01 10 deka da
    1e+00 1
    1e-01 0.1 deci d
    1e-02 0.01 centi c
    1e-03 0.001 milli m
    1e-06 0.000 001 micro
    1e-09 0.000 000 001 nano n
    1e-12 0.000 000 000 001 pico p
    1e-15 0.000 000 000 000 001 femto f
    1e-18 0.000 000 000 000 000 001 atto a
    1e-21 0.000 000 000 000 000 000 001 zepto z
    1e-24 0.000 000 000 000 000 000 000 001 yocto y

    Color issue #1 (by Blizzard):
    Hold LMB or RMB but do not rotate the camera, just move your character
    The cursor at this time should not disappear from the screen, it will follow the movement of your character
    Thus, while it follows, when hovering over a unit there is a color issue
    Because unit returns nil under the given conditions

    Color issue #2:
    Hover over a unit (during combat) and then move the cursor away from him
    The tooltip should not disappear from the screen at this time (FadeOut animation)
    Then quickly let the unit take some damage to see the changes
    The status bar color will turn to default green, not class color
    Because unit returns nil under the given conditions
]]

local cache = {}

local function abbreviateLargeNumbers(value)
    if value >= 1e6 then return string.format('%.1fM', value / 1e6) -- 1.6M, 61.8M
    elseif value >= 1e5 then return string.format('%dk', value / 1e3) -- 128k
    elseif value >= 1e3 then return string.format('%.1fk', value / 1e3) -- 1.6k, 61.8k
    else return value
    end
end

local function addLine(tooltip, targetLine, text, color)
    local color = color or CreateColor(1, 1, 1)

    -- Text and color do not matter at this step
    tooltip:AddLine(text, 1, 1, 1)
    tooltip:Show() -- update borders correctly

    for i = tooltip:NumLines(), 2, -1 do
        local currLine = _G[tooltip:GetName() .. 'TextLeft' .. i]
        local nextLine = _G[tooltip:GetName() .. 'TextLeft' .. i - 1]

        -- Find the desired line position by finding the next line
        if nextLine == targetLine then
            currLine:SetText(text)
            currLine:SetTextColor(color.r, color.g, color.b)
            return -- stop the loop and exit the function

        -- If the next line does not meet our requirements, then copy it to the current one
        -- The RESURRECTABLE constant is green, so copy the color as well
        else
            currLine:SetText( nextLine:GetText() )
            currLine:SetTextColor( nextLine:GetTextColor() )
        end
    end
end

-- Fix the color issue #2 described above
local function getUnitData(unit)
    -- The issue also occurs when using UnitPlayerControlled()
    -- Nameplates must be turned off to see this
    cache.prevUnit = cache.currUnit
    cache.currUnit = UnitPlayerControlled(unit)

    if UnitIsPlayer(unit) then
        local class = select(2, UnitClass(unit))
        local classColor = RAID_CLASS_COLORS[class]

        cache.prevColor = cache.currColor
        cache.currColor = classColor
    else
        cache.prevColor = cache.currColor
        cache.currColor = CreateColor(0, 1, 0)
    end

    if unit == nil then
        cache.currUnit = cache.prevUnit
        cache.currColor = cache.prevColor
    end
    return cache.currUnit, cache.currColor
end

local function setStatusBar(statusBar, unit)
    local unitIsPlayer, unitColor = getUnitData(unit)
    local unitIsDead = unit and UnitIsDeadOrGhost(unit) -- no need to save state in cache
    local currHealth = statusBar:GetValue()
    local maxHealth = select(2, statusBar:GetMinMaxValues())

    -- ChatFrame1:Clear()
    -- print( 'unitIsPlayer = ' .. tostring(unitIsPlayer) )
    -- print( 'unitIsDead = ' .. tostring(unitIsDead) )
    -- print( 'unitColor = ' .. string.format('%.1f, %.1f, %.1f', unitColor.r, unitColor.g, unitColor.b) )

    if currHealth <= 0 or unitIsDead then
        cache.healthText = DEAD
    elseif unitIsPlayer then
        cache.healthText = math.ceil(currHealth/maxHealth * 100) .. '%' -- enemy players or the units they summon
    else
        cache.healthText = abbreviateLargeNumbers(currHealth) .. ' / ' .. abbreviateLargeNumbers(maxHealth) -- NPCs
    end
    statusBar.healthText:SetText(cache.healthText)
    statusBar:SetStatusBarColor(unitColor.r, unitColor.g, unitColor.b)
end

local function onTooltipSetUnit(self)
    if IsMouseButtonDown('LeftButton') or IsMouseButtonDown('RightButton') then
        return self:Hide() -- fix the color issue #1 described above
    end

    local unit = select(2, self:GetUnit())
    local statusBar = self:GetChildren()
    cache.currUnit = UnitPlayerControlled(unit)

    -- ChatFrame1:Clear()
    -- print( 'unitType = ' .. tostring(unit and UnitGUID(unit)):lower():match('%l+') )

    if UnitIsPlayer(unit) then
        local unitName = GetUnitName(unit, true) -- ignore PvP title
        local guildName = GetGuildInfo(unit)
        local class = select(2, UnitClass(unit))
        local classColor = RAID_CLASS_COLORS[class]
        local playerFaction = select(2, UnitFactionGroup('player'))
        local unitFaction = select(2, UnitFactionGroup(unit))

        for i = 1, self:NumLines() do
            local line = _G[self:GetName() .. 'TextLeft' .. i]

            -- Set class color
            if i == 1 then
                line:SetText(
                    '|c' .. classColor.colorStr .. unitName .. '|r' -- fix UnitFrame name color
                )
                cache.currColor = classColor

            -- Set guild line color
            elseif i == 2 and guildName then
                line:SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)

            -- There will always be the level line at this step
            else
                cache.levelLine = line
                break
            end
        end

        -- Add faction line
        addLine(
            self, cache.levelLine, unitFaction,
            -- Set faction line color, if the player is from the opposite faction
            -- Convert RGB(%) to Hex: string.format('|cff%02x%02x%02x%s|r', r * 255, g * 255, b * 255, text)
            playerFaction ~= unitFaction and FACTION_BAR_COLORS[1]
        )
    else
        -- Revert to default color, if the unit is not a player (or nil)
        cache.currColor = CreateColor(0, 1, 0)
    end
    setStatusBar(statusBar, unit)
end

local function onValueChanged(self, value)
    local unit = select(2, self:GetParent():GetUnit())

    setStatusBar(self, unit)
end

GameTooltipStatusBar.healthText = GameTooltipStatusBar:CreateFontString(nil, 'OVERLAY')
GameTooltipStatusBar.healthText:SetFont(STANDARD_TEXT_FONT, 13, 'OUTLINE') -- 'Fonts\\ARIALN.TTF'
GameTooltipStatusBar.healthText:SetPoint('CENTER')

GameTooltip:HookScript('OnTooltipSetUnit', onTooltipSetUnit)
GameTooltipStatusBar:HookScript('OnValueChanged', onValueChanged)